﻿using Microsoft.AspNetCore.Authentication;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Options;
using Netmash.Security.Authentication.Tuitio.Abstractions;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net.Http.Headers;
using System.Security.Claims;
using System.Text.Encodings.Web;
using System.Threading.Tasks;
using Tuitio.PublishedLanguage.Dto;
using Tuitio.Wrapper.Services;
using c = Netmash.Security.Authentication.Tuitio.Constants;

namespace Netmash.Security.Authentication.Tuitio
{
    public class TuitioAuthenticationHandler : AuthenticationHandler<AuthenticationSchemeOptions>
    {
        private readonly ITuitioService _tuitioService;
        private readonly IAuthenticationOptions _authenticationOptions;
        private readonly ILogger<TuitioAuthenticationHandler> _logger;

        public TuitioAuthenticationHandler(IOptionsMonitor<AuthenticationSchemeOptions> options, ILoggerFactory loggerFactory, UrlEncoder encoder, ISystemClock clock, ITuitioService tuitioService, IAuthenticationOptions authenticationOptions, ILogger<TuitioAuthenticationHandler> logger)
            : base(options, loggerFactory, encoder, clock)
        {
            _tuitioService = tuitioService;
            _authenticationOptions = authenticationOptions;
            _logger = logger;
        }

        protected override async Task<AuthenticateResult> HandleAuthenticateAsync()
        {
            var token = GetAuthorizationToken();
            if (token != null)
            {
                Envelope<AuthorizationResult> authorizationEnvelope;
                try
                {
                    authorizationEnvelope = await _tuitioService.Authorize(token);
                }
                catch (Exception e)
                {
                    _logger.LogError(e, "Tuitio authorization failed");
                    return AuthenticateResult.Fail("UNAUTHORIZED");
                }

                if (!string.IsNullOrEmpty(authorizationEnvelope.Error))
                    return AuthenticateResult.Fail(authorizationEnvelope.Error);

                var ticket = GetAuthenticationTicket(authorizationEnvelope.Result);
                return AuthenticateResult.Success(ticket);
            }

            var authenticateAsGuest = _authenticationOptions.AuthenticateAsGuest?.Invoke(Request) ?? false;
            if (authenticateAsGuest)
            {
                var guestTicket = GetGuestAuthenticationTicket(_authenticationOptions.GuestUserId, _authenticationOptions.GuestUserName);
                return AuthenticateResult.Success(guestTicket);
            }

            return AuthenticateResult.Fail("AUTHORIZATION_HEADER_IS_MISSING");
        }

        private string GetAuthorizationToken()
        {
            if (Request.Headers.ContainsKey("Authorization"))
            {
                var authorizationHeader = AuthenticationHeaderValue.Parse(Request.Headers["Authorization"]);
                var token = authorizationHeader.Parameter;
                return token;
            }

            if (_authenticationOptions.AcceptTokenFromQuery
                && Request.Query.Count > 0
                && Request.Query.ContainsKey(c.QueryParams.Token))
            {
                var token = Request.Query[c.QueryParams.Token];
                return token.ToString();
            }

            return null;
        }

        private AuthenticationTicket GetGuestAuthenticationTicket(int guestId, string guestName)
        {
            var claims = new[] {
                new Claim(ClaimTypes.NameIdentifier, guestId.ToString()),
                new Claim(ClaimTypes.Name, guestName),
                new Claim(Constants.ClaimTypes.IsGuestUser, bool.TrueString)
            };

            var ticket = GetAuthenticationTicket(claims);
            return ticket;
        }

        private AuthenticationTicket GetAuthenticationTicket(AuthorizationResult authorization)
        {
            var claimCollection = new Dictionary<string, string>()
            {
                { ClaimTypes.NameIdentifier, authorization.UserId.ToString() },
                { ClaimTypes.Name, authorization.UserName },
                { Constants.ClaimTypes.UserName, authorization.UserName }
            };

            if (authorization.FirstName != null)
            {
                claimCollection.Add(ClaimTypes.GivenName, authorization.FirstName);
                claimCollection.Add(Constants.ClaimTypes.FirstName, authorization.FirstName);
            }

            if (authorization.LastName != null)
            {
                claimCollection.Add(ClaimTypes.Surname, authorization.FirstName);
                claimCollection.Add(Constants.ClaimTypes.LastName, authorization.LastName);
            }

            if (authorization.Email != null)
                claimCollection.Add(ClaimTypes.Email, authorization.Email);

            if (authorization.Claims != null && authorization.Claims.Any())
            {
                foreach (var claim in authorization.Claims)
                {
                    if (claimCollection.ContainsKey(claim.Key))
                    {
                        _logger.LogWarning($"There is already a claim with key {claim.Key} in the collection. The combination {claim.Key}:{claim.Value} will be ignored.");
                        continue;
                    }
                    claimCollection.Add(claim.Key, claim.Value);
                }
            }

            var claims = claimCollection.Select(z => new Claim(z.Key, z.Value)).ToArray();
            var ticket = GetAuthenticationTicket(claims);

            return ticket;
        }

        private AuthenticationTicket GetAuthenticationTicket(Claim[] claims)
        {
            var identity = new ClaimsIdentity(claims, Scheme.Name);
            var principal = new ClaimsPrincipal(identity);
            var ticket = new AuthenticationTicket(principal, Scheme.Name);

            return ticket;
        }
    }
}
